
/*
 * Multi-Timer
 * (c) 2019  J. Muller
 * 
 * Supports a Multi-timer in hardware that provides
 * timing with Alarm for:
 * 
 * 1) Napping for  a fixed set of tomes between 20 minutes
 *    and 3 hours
 *    
 * 2) Meditation with a fixed set of times of 20, 30, 35 
 *    and 40 minutes with a short chirp alarm at end of 
 *    preiod and an additional chirp at end plus 5 minutes
 *    
 * 3) General purpose timing( Standard) user selectable times 
 *    in two ranges (1 second  - 60 minutes or 1 minute to 
 *    99 hours.)  This option will  provide an audible
 *    alarm and/or control of a jumper selectable voltage source. 
 *    The voltage source may be internal (6 V output ) or external.  
 *    This allows the timer  to control a low voltage external device
 *    such as a UV exposure box.
 *    
 *    The hardware design and build are covered in associated 
 *    The instructable project .   
 *    
 *    E N H A N C M E N T S
 *    
 *    For those who wish to enhance this code the following tasks
 *    are suggested, there is plenty of memory left (about 1/2) for this
 *    
 *    1) Prevent the double meditation alarm tones when canceling
 *       meditation time when pressing START
 *       
 *    2) Provide repeat time ability in the Standard Timer
 *       to avoid re-inputing the time value.
 *       
 *    3) Add a stop watch type with pause while encoder pressed.
 *       stops timer and display results when Start pressed and end
 *       when pressed a sedond time
 *    
 *    4) remove repeated code by placeing inside new functions 
 *    
 *    
 */ 
#include <RotaryEncoder.h>

#include <U8g2lib.h>  // For OLED
#include <U8x8lib.h>

#include <Bounce2.h>

/*
 *  Arduino I/O
*/
// Rotary Encoder
#define ROTEN_A  2
#define ROTEN_B  3
#define ROTEN_PB 4

// OLED uses standard Arduino I2C pins A4 & A5

// Start switch, speaker & control relay pins
#define START 10
#define SPKER 11
#define RELAY 6

/*
 *   MACROS & Standard Values
 */
#define forever for(;;)  // infinite loop
#define SELIS  F("Selection is: ")
#define MILPERSEC 1000   // number of milseconds per second
#define SECPERMIN 60 
#define SECPERHR SECPERMIN * SECPERMIN

#define RLYON  HIGH
#define RLYOFF LOW

//#define AUDIBLE 0X01
//#define CONTROL 0X02

/*
 *   Program Constants & Enumerations
*/
 
enum STATES {  // possible values for the Finite State Machine
   ERR = -1, INITIAL = 0, SELTYPE, NAPSEL, NAPST, NAPTM,
   NAPALM, NAPOVR, MEDSEL, MEDST, MEDTM, MEDALM, MEDRCV, 
   STDRANGE, STDSEL, STDCTL, STDST, STDTM, STDALM, STDOVR
};

const byte AUDIBLE = 0X01;
const byte CONTROL = 0X02;

// de-bounce time in miliseconds adjust as needed
const int PBDB = 25;  // Encoder push button 
const int STDB = 25;  // Start push button
/*
 *  Global Variables
*/

STATES state = ERR;  // current state

long timerTarget;
int timerValue;

int timerVal;  // count up / down timer value in seconds
bool minsecfmt;
byte alarmType = 0;  //  none, audible, control, both

// Instantiate Classes
Bounce debouncerR = Bounce(); 
Bounce debouncerS = Bounce(); 

// OLED
U8X8_SSD1306_128X64_NONAME_HW_I2C u8x8(/* reset=*/ U8X8_PIN_NONE);    

void setup() {
  // Define I/O
  pinMode( START, INPUT_PULLUP);
  pinMode( SPKER, OUTPUT);
  pinMode( RELAY, OUTPUT);
  
  digitalWrite( RELAY, RLYOFF); // set Relay output high to avoid triggering @ startup
  
  // rotary encoder
  pinMode( ROTEN_PB, INPUT_PULLUP);
  pinMode( ROTEN_A, INPUT_PULLUP);
  pinMode( ROTEN_B, INPUT_PULLUP);
  
  // setup Serial port for testing & debug
  Serial.begin(115200);
  delay(5000);  // Switch to USB delay after loading code

  Serial.println(F("Starting Program"));  // NOTE use of program
                              // memory for print statements

  // Debounce settings for encoder & start buttons 
  // 
  debouncerR.attach(ROTEN_PB);
  debouncerR.interval(PBDB); 

  debouncerS.attach(START);
  debouncerS.interval(STDB); 
 
  oledInit();  // setup up the OLED 
  
  state = INITIAL;  
} // setup()


void loop() {

   // NOTE:  Any variables used inside a switch
   // block that need to be retained across states
   // must be defined as globals NOT inside loop function
     
   int selection = -1; // the value of the selection made
   int cntTimer = 0;  // used by 1 second timer loop (blocking)
   int exitVal;  // tests function exits for reason

   const int DISPLAY_SEC = 30; // * seconds to show screen before off
   // The Finite State Machine is encoded by this switch statement
   switch(state){
      // Display the Application Splash screen till start pressed 
      // or 30 seconds elapsed
      case INITIAL:
         Serial.println(F("State: INITIAL"));
         displayIntital();

         setTimer(MILPERSEC);  // set elapsed time (ms)
         while( ! isStartPressed()) {
            // call timer true if delay time
            if( timeTick()){ //
               cntTimer++;
            }

            // test if delay time exceeded
            if( cntTimer >= DISPLAY_SEC) {
               state = ERR;
               return;  // exits loop & process new state
            }
         }
         
         state = SELTYPE; // next state
         break;  // case INITIAL

      // Select the type of timing desired
      case SELTYPE:
         Serial.println(F("State: SELTYPE"));
         selection = displaySeltype();        
//         Serial.print(SELIS);   // for debug
//         Serial.println(selection);

         switch( selection) {
            case 0:
               state = NAPSEL;
               break;

            case 1:
               state = MEDSEL;
               break;

            case 2:
               state = STDRANGE;
               break;
                    
            default:  // start button pressed 
              state = INITIAL;
              break;
         }  // switch selection
         
         break;  // case SELTYPE

      // How long to nap?
      case NAPSEL:
         Serial.println(F("State: NAPSEL"));
         selection = displayNapSelect();
         //Serial.print(SELIS);   // for debug
         //Serial.println(selection); 

         if(selection < 0) {  // abort on start pressed
            state = SELTYPE;
            return;  // restart loop
         }
         switch(selection) {
            case 0:
               timerVal = 20 * SECPERMIN;
               //timerVal = 20;  // TESTING comment above
               minsecfmt = true;
               break;

            case 1:
               timerVal = 30 * SECPERMIN;
               minsecfmt = true;
               break;
               
            case 2:
               timerVal = 40 * SECPERMIN;
               minsecfmt = true;
               break;

            case 3:
               timerVal = 60 * SECPERMIN;
                minsecfmt = false;
                break;
 
            case 4:
               timerVal = 90 * SECPERMIN;
               minsecfmt = false;
               break;

            case 5:
               timerVal = 2 * SECPERHR;
               minsecfmt = false;
               break;

            case 6:
               timerVal = 3 * SECPERHR;
               minsecfmt = false;
               break;
               
            default:  // try again at start!
               state = SELTYPE;
               break;   
         }
         
         state = NAPST;
         break;  // case NAPSEL

      // display count down time (timerVal)
      // wait for start button pressed to start nap timing
      case NAPST:
         Serial.println(F("State: NAPST"));
          
         displayNapTime();     
         state = NAPTM;
         break;  // case NAPST

      // Display nap time countdown till alarm at 0
      case NAPTM:
         Serial.println(F("State: NAPTM"));
         
         displayNapCountdn();
         state = NAPALM;
         break;  // case NAPTM

      // Sound the nap alarm & count up till start pressed
      // max alarm time 5 min
      case NAPALM:
         Serial.println(F("State: NAPALM"));
         
         displayNapAlarm();
         state = NAPOVR;
         break;  // case NAPALM

      // clean up from nap time - return to selection menu
      // This state can be used in future enhancements
      case NAPOVR:
         Serial.println(F("State: NAPOVR"));

         // no need for display screen just turn off alsrm and 
         // reset display to normal
         noTone(SPKER);
         u8x8.setInverseFont(0);
         
         state = SELTYPE;
         break;  // case NAPOVR

      // Select the meditation time desired
      case MEDSEL:
         Serial.println(F("State: MEDSEL"));
         selection = displayMedSelect();
         //Serial.print(SELIS);   // for debug
         //Serial.println(selection);

         if(selection < 0) {  // abort - start pressed
            state = SELTYPE;
            return;  // restart loop
         }
         switch(selection) {
            case 0:
               timerVal = 20 * SECPERMIN;
               //timerVal = 20;  // TESTING comment above
               minsecfmt = true;
               break;

            case 1:
               timerVal = 30 * SECPERMIN;
               minsecfmt = true;
               break;
               
            case 2:
               timerVal = 35 * SECPERMIN;
               minsecfmt = true;
               break;

            case 3:
               timerVal = 40 * SECPERMIN;
                minsecfmt = true;
                break;
               
            default:  // Invalid selection DEBUG!
               state = ERR;
               break;   
         }
         state = MEDST;
         break;  // case MEDSEL

      // Show selected time and wait for start pressed
      case MEDST:
         Serial.println(F("State: MEDST"));

         displayMedTime(); 
         state = MEDTM;
         break;  // case MEDST

      // meditation time count down & display till 0
      case MEDTM:
         Serial.println(F("State: MEDTM"));

         displayMedCountdn();
         state = MEDALM;
         break;  // case MEDTM

      // sound med done chirp 
      case MEDALM:
         Serial.println(F("State: MEDALM"));

         displayMedAlarm();
         state = MEDRCV;
         break; // case MEDALM

      // meditation recovery time & count up till + 5 min then chirp 
      // once wait till start pressed 
      case MEDRCV:
         Serial.println(F("State: MEDRCV"));

         displayMedRecover();
         u8x8.setInverseFont(0);
         state = SELTYPE;
         break;  // case MEDRCV
         
      // select timer range Hr/min or min/sec
      case STDRANGE:
         Serial.println(F("State: STDRANGE"));

         selection = displayStdRange();
         if(selection < 0) {  // abort on start pressed
            state = SELTYPE;
            return;  // restart loop
         }
         switch(selection) {
            case 0:
               minsecfmt = true;
               break;

            case 1:
               minsecfmt = false;
               break;
               
            default:  // try again at start!
               state = SELTYPE;
               break;   
         }         
         state = STDCTL;
         break;  // case STDRANGE

      // select control and or alarm types
      case STDCTL:
         Serial.println(F("State: STDCTL"));
         
         alarmType = displayStdCtl();  // assigns global value
         state = STDSEL;
         break;  // case STDCTL
         
      // Select time period (free form HH:MM or MM:SS)
      case STDSEL:
         Serial.println(F("State: STDSEL"));

         displayStdSel();  // sets global timerVal variable
         state = STDST;
         break;  // case STDSEL

      // show the input time and wait for Srart button pressed
      case STDST:
         Serial.println(F("State: STDST"));

         displayStdTime(); 
         state = STDTM;
         break;  // case STDST

      // start counting down time exit on start pressed (abort) or
      // time = zero
      case STDTM:
         Serial.println(F("State: STDTM"));

         exitVal = displayStdCountdn();
         Serial.println(exitVal);
         if(exitVal < 0) { // abort
            state = SELTYPE;
            digitalWrite(RELAY, RLYOFF); // off if aborted
            return;  // restart loop
         }
         state = STDALM;    
         break;  // case STDTM

      // Sounds Alarm (if audible) turns off the relay and shows
      // timer count up exit when start button pressed
      case STDALM:
         Serial.println(F("State: STDALM"));

         displayStdAlarm();
         state = STDOVR;
         break;  // case STDALM

      // Turns off the alarm and changes state to the timer 
      // type selection
      case STDOVR:
         Serial.println(F("State: STDOVR"));

         noTone(SPKER);  //always turn off even if already off
         state = SELTYPE;
         break;  // case STDOVR

      // Severe error or app timeout turn off OLED to conserve battery
      // power and wait till user turns the switch off.
      case ERR:
         Serial.println(F("Un-recoverable Error"));
         OLEDoff();
         forever;
         break;
         
      default:  // unknown assignment to state 
         state = ERR; // show unrecoverable error message
         break;
   }  // switch(state)

}  // loop()

// sets up the reset time for the tic timer below
void setTimer(int value) {
   timerTarget = value + millis(); // first timing
   timerValue = value;  // reset value in ms
}

//  This is non blocking timer but do not use for critical timing
//  since no adjustment is made for processing time within
//  this code.
bool timeTick(){  
   // Serial.println( millis(), timerTarget);
   if( millis() > timerTarget) {
      timerTarget = timerValue + millis(); // reset timer
      return true;  // tic the timer 
   }
   return false;  // no tic
}
