/*
 *    Code used to control the OLED Display   
 */

#define MAXSEL 10
#define MAXPOS 14
#define BLANK  F("                ")
#define FLIP 1  // set to 1 for flip 0 for no flip


// Defines information about menu selection options
struct cursorpos {
  uint8_t line;  // menu line (row) starts at 1
  uint8_t startpos; //  menu item x position
  char text[MAXSEL]; // null terminated menu item text to display
};


/*
 *  This is a custom class to display a series of text as a
 *  selection menu interfaces with the rotary encoder and
 *  encoder push button to allow users to select a menu option
 *  Uses u8x8 library
 */
class Menu {
  public:
    Menu() { };   // Construct this object once at setup


    // initial is called for each new memu to display
    // it resets the class variables
    void initial(char* title, int itemcnt, int defitem = 0, int start = 3) {
      menucnt = itemcnt;
      menupick = defitem;
      menuStartLine = start;

      setupEnc( itemcnt, menupick);  // Note this could also be set in
                                    // the code that calls this method 
                                    // to override
      // clear out positions   NEEDED??
      for ( int i = 0; i < itemcnt; i++) {
        _positions[i].line = -1;
      }
      curpos = 0;
      // title always on second line
      u8x8.setCursor(0, 1);
      u8x8.print( title);    

      //oled.display();
    };  // initial()


    // adds a menu selection item to the menu
    // displayed at a line and a column (col)
    // using a zero terminated string (t)
    void addMenuSel(int line, int col, char* t ) {
      _positions[curpos].line = line;
      _positions[curpos].startpos = col;

      for( int i = 0; i < MAXSEL; i++) {
         _positions[curpos].text[i] = t[i]; // copy string t 
         if ( t[i] == '\0' ) break;  // exit at end of string       
      }
      
      _positions[curpos].text[MAXSEL -1] = '\0';  // force null if at max size
      curpos++;  // next menu selection
      if ( curpos >= MAXPOS){
         Serial.println(F("Menu items Size exceeded"));
         curpos = MAXPOS - 1;
      }

    };  // addMenuSel()


   // display the selections for this Menu 
   // handles highlighting
    void showMenu() { 
      int menuline = 0;
      // remove all menu option lines
      for( int i = 0; i < menucnt; i++) {  // clear menu lines
         if( _positions[i].line != menuline) {
            // clear the line
            u8x8.setCursor(0,  _positions[i].line + menuStartLine);
            u8x8.print(BLANK);
            menuline = _positions[i].line;
         }
      }  // for clear lines

      // place menu items selected item inverted
      for( int i = 0; i < menucnt; i++) {
//         Serial.println(menupick);
         if( i == menupick) {
            u8x8.setInverseFont(1);  // invert menu pick
         }
         else {
            u8x8.setInverseFont(0);  // others normal
         }

         // position print start location col & row
         u8x8.setCursor(_positions[i].startpos, _positions[i].line + menuStartLine);
         u8x8.print(_positions[i].text);       

      }  // for menu item display
      
      u8x8.setInverseFont(0);  // reset to normal in case selection is last
      
    }; // showmenu()


    int  menupick;  // public variables are visible outside of class
    int  menucnt;
    int  menuStartLine;

  private:  // can only be used inside class functions (methods)
    cursorpos _positions[MAXPOS];
    int curpos;
};  // class Menu

Menu myMenu = Menu();  // instantiate a single Menu object used throughout the program


/*************************************************************
 *  Button tests
 **************************************************************/
 
bool isEncPressed() {
   debouncerR.update();
   if( debouncerR.read() == LOW) {
      // wait till released & return true
      do {
         debouncerR.update();
      } while(debouncerR.read() == LOW);
      return true;
   }
   // not pressed 
   return false;
}

bool isStartPressed() {
   debouncerS.update();
   if( debouncerS.read() == LOW) {
      // wait till released & return true
      do {
         debouncerS.update();
      } while(debouncerS.read() == LOW);
      return true;
   }
   // not pressed 
   return false;
}

/*************************************************************
 *   Display helper functions
 ************************************************************/
// App oled heading
void startDisplay() {  
   u8x8.clear(); // clear and home
   u8x8.print(F("  Multi-Timer   "));
}

 
void oledInit(){   // called once from setup
  u8x8.begin();
  u8x8.setFlipMode(FLIP);  
  u8x8.setPowerSave(0); 
   
  // use a Font (change as desired but TEST)
  u8x8.setFont(u8x8_font_chroma48medium8_r);   
  u8x8.drawString(0,0,"Initial Test");  // self test DEBUG
}

void OLEDoff() {  // Turn off to save power
   u8x8.clear();
   u8x8.setPowerSave(1);
}

   
/*
 *   OLED display Functions
 */

// General function handle menu selection after setup
// does not return till selection made or start pressed
// returns menu selection or -1 if start button
 int getMenuSel() {
   int rval = -99;  // force a change first time 
   // loop till button pressed
   while ( true) {  

      // if rotary button pressed exit with menu pick
      if( isEncPressed())  return rval;
      
      // if start button pressed return -1
      if( isStartPressed()) return -1;  
      
      // get rotary encoder 
      int v = getEncValue();
      if(  v != rval) {
//         Serial.print("Enc changed ");
//         Serial.println(v);

         // update menu selection and redisplay menu 
         rval = v;
         myMenu.menupick = v;
         myMenu.showMenu(); 
      }
    }  // while
   
 } //  getMenuSel()

// displays the contents of timerVal as "HH:MM" or "MM:SS"
// format 2x2 font starting at line
void showTime(bool mmss, int line) {
   char major[2], minor[2];
   char out[7];
   char rng;
   int tval;

   // convert timeVal HH:MM or MM:SS
   // note timeVal always in seconds
   if( mmss) {
      tval = timerVal; // leave as seconds
      rng = 's';
   }
   else {
      tval = timerVal / SECPERMIN;  // convert to minutes
      rng = 'm';
   }
   Serial.println(timerVal);
   Serial.println(tval);
   
   int small = tval % SECPERMIN;
   int large = tval / SECPERMIN;

   if (large > SECPERHR ) {
   Serial.println("To large for mm:ss format");
   forever;
   }

  // handle large value
  if (large < 10) {
    // need to add a zero to major
    major[0] = '0';      
  }
  else {
    major[0] = (char) ((large / 10) + '0');
  }
  
  major[1] = (char) (large % 10 + '0'); // adding to char '0' converts to ascii digit
  
  // handle small value
  if( small < 10) {
   minor[0] = '0';
  }
  else{
   minor[0] = (char) ((small / 10) + '0');
  }

  minor[1] = (char) ((small % 10) + '0');
   
   
   // put it together and display
   out[0] = major[0];
   out[1] = major[1];
   out[2] = ':';
   out[3] = minor[0];
   out[4] = minor[1];
   out[5] = rng;
   out[6] = '\0';  // null termiator
   //Serial.println(out);  // DEBUG
   u8x8.draw2x2String(2, line, out);    

}  // showTime()

// displays and allow selection of time major and minor values
// 
// A lot of time was sent writing and debugging this function
// as a consequence it contains the least elegant code
// a more skilled coder could re-write it, but it works
// 
int getFreeTime() {
   int inTime = 0;
   bool inMajor = true;
   int rval = 0;
   int tsec, tmin;

   // Had to add following line to set up encoder before initial
   // encoder pb press 
   showFreeTime(inTime, inMajor);

   // still an issure as initially can not
   // decrement counter below zero and get 99 as major value
   //This is an artifact of the getEncValue function!
   // it does not allow minus values,unless the internal encoder value 
   // it is > max it will not work correctly. 
   // if this is critical for you may re-write the getEncValue code to correct  
   setupEnc(100, inTime / SECPERMIN);  // assume major time
   while( ! isStartPressed()) { // till start pressed
      tsec = inTime % SECPERMIN;  // minor time period
      tmin = inTime / SECPERMIN;  // major time period
      
      if( isEncPressed()) {
         Serial.println("enc press");
         rval = -1; // force update change
         inMajor = (! inMajor);
         showFreeTime(inTime, inMajor);
         tsec = inTime % SECPERMIN;  // minor time period
         tmin = inTime / SECPERMIN; 
         // set up encoder
         if( inMajor) {
            setupEnc( 100, tmin);
         }
         else { // minor
            setupEnc( SECPERMIN, tsec);
         }
         showFreeTime(inTime, inMajor);
      }
      // look for change
      int v = getEncValue();
      if ( rval != v) {
//         Serial.println(tmin);  // for testing
//         Serial.println(tsec);
         // we have a change
         rval = v;

         
         if( inMajor) {
           inTime = (v * SECPERMIN) + tsec; // Note: I try to always use parens
                                        // even if not needed to make the
                                        // calculation intention clear 
         }
         else {
           inTime = (tmin * SECPERMIN) + v;
         }
         // update changes
        // display input time value only if changed
         showFreeTime(inTime, inMajor);
      }  // value change
 
   } // while start button not pressed

   return (tmin * SECPERMIN) + tsec;
}  // getFreeTime

void showFreeTime( int dtime, bool majinv) {
   char major[3], minor[3];
   const int line = 4;
 
   // convert input time to Major & Minor test values
   int small = dtime % SECPERMIN;
   int large = dtime / SECPERMIN;

   if (large < 10) {
    // need to add a zero to major
    major[0] = '0';      
  }
  else {
    major[0] = (char) ((large / 10) + '0');
  }
  
  major[1] = (char) (large % 10 + '0'); // convert to ascii
  
  // handle small
  if( small < 10) {
   minor[0] = '0';
  }
  else{
   minor[0] = (char) ((small / 10) + '0');
  }

  minor[1] = (char) ((small % 10) + '0');

  major[2] = '\0'; // null terminators
  minor[2] = '\0';    
  //
  if ( majinv) { // major value highlighted
      u8x8.setInverseFont(1); 
      u8x8.draw2x2String(2, line, major);
      u8x8.setInverseFont(0); 
      u8x8.draw2x2String(6, line, ":");
      u8x8.draw2x2String(7, line, minor);
  }

  else {  // minor value highlighted      
      u8x8.draw2x2String(2, line, major);       
      u8x8.draw2x2String(6, line, ":");
      u8x8.setInverseFont(1); 
      u8x8.draw2x2String(7, line, minor);
      u8x8.setInverseFont(0);
  }

  u8x8.setInverseFont(0); // insurance
}   // showFreeTime()


/********************************************************
 *   Functions that display the State screens
 ********************************************************/
 
// called from INITIAL state
void displayIntital() {
   startDisplay();
   u8x8.setCursor(0, 2);
   u8x8.print(F("(c)2019 J.Muller"));  // change this if you modify the code
   u8x8.setCursor(0, 3);
   u8x8.print(F("   Version 2.0  "));  // also increment the Version
   u8x8.setCursor(0, 4);
   u8x8.print(F("   Press START  "));
}  //  displayIntital()


// menu selection state returns menu sel when encoder pb selected
int displaySeltype() {
   int rval;
   
   startDisplay();
   u8x8.setCursor(0,2);  // set position of menu title
   myMenu.initial("    Timer Type", 3, 1, 3); // menu title, # menu items, initial selected
                                              // starting at line
   myMenu.addMenuSel(0, 3, "Nap" );  // line, col, menu text
   myMenu.addMenuSel(1, 3, "Meditate" ); 
   myMenu.addMenuSel(2, 3, "Stardard" ); 
   myMenu.showMenu(); // show items 

   return getMenuSel();
   
}  // displaySeltype()


// allows selection of meditation periods
int displayNapSelect() {

   startDisplay();
   u8x8.setCursor(0,2);  // set position of menu title
   myMenu.initial("   Nap Length ", 7, 1, 3); // menu title, # menu items, initial selected
                                              // starting at line
   myMenu.addMenuSel(0, 1, "20m" );  // line, col, menu text
   myMenu.addMenuSel(0, 6, "30m" ); 
   myMenu.addMenuSel(0, 11, "40m" );
   myMenu.addMenuSel(1, 1, "60m" );
   myMenu.addMenuSel(1, 6, "90m" );
   myMenu.addMenuSel(2, 1, "2h" );
   myMenu.addMenuSel(2, 6, "3h" ); 
   myMenu.showMenu(); // show items    

   return getMenuSel();
   
 } //displayNapLength()

// displays the selected nap time waits for start press
int displayNapTime() {
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print(" Selected Time");
    showTime(minsecfmt, 4);
    u8x8.setCursor(0,7);  // show start msg
    u8x8.print("Press START ->");

    // wait for start pressed
    while(! isStartPressed()) {
      
    }
    
}  //  displayNapTime()

// shows the nap timer counting down
// exit on zero time
void displayNapCountdn() {
    long targetTm;
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print("   Nap Timer    ");
    showTime(minsecfmt, 4);
    targetTm = millis() + 1000;
     while( true) {
         if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal--;
            if( timerVal <= 0) return;
            // test for minute update
            if( ! minsecfmt && (timerVal % 60) != 0) continue;  // while loop 

            // show the time
            showTime(minsecfmt, 4);
         }
      
     }  // while
         
}  // displayNapCountdn()

// sound the nap alarm, count up & wait for start press to exit
void displayNapAlarm() {
   
    tone(SPKER, 440);
    long targetTm;
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print("   Nap Over    ");
    u8x8.setInverseFont(1);
    showTime(minsecfmt, 4);
    targetTm = millis() + 1000;

   while ( ! isStartPressed()) {
      // process the time increment timerVal & display 1 sec intervals
      if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal++;
          
            // show the time
            showTime(minsecfmt, 4);
       }
   }  // while

   // alarm end

}  //displayNapAlarm()

int displayMedSelect() {
   startDisplay();
   u8x8.setCursor(0,2);  // set position of menu title
   myMenu.initial("Meditation Time ?", 4, 1, 3); // menu title, # menu items, initial selected
                                              // starting at line
   myMenu.addMenuSel(0, 3, "20m" );  // line, col, menu text
   myMenu.addMenuSel(0, 9, "30m" ); 
   myMenu.addMenuSel(1, 3, "35m" );
   myMenu.addMenuSel(1, 9, "40m" );
  
   myMenu.showMenu(); // show items    

   return getMenuSel();  
}  //  displayMedSelect()

void displayMedTime() {
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print(" Selected Time");
    showTime(minsecfmt, 4);
    u8x8.setCursor(0,7);  // show start msg
    u8x8.print("Press START ->");

    // wait for start pressed
    while(! isStartPressed()) {
      
    }
}  // displayMedTime()

void displayMedCountdn() {
   long targetTm;
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print("Meditation Timer");
    showTime(minsecfmt, 4);
    targetTm = millis() + 1000;
     while( ! isStartPressed()) {
         if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal--;
            if( timerVal <= 0) return;

            // show the time
            showTime(minsecfmt, 4);
         }
      
     }  // while   
}  // displayMedCountdn()

void displayMedAlarm() {
    tone(SPKER, 220, 500);  // 1/2 sec chirp
    long targetTm;

    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print("Meditation Over ");
    u8x8.setInverseFont(1);
    showTime(minsecfmt, 4);
    targetTm = millis() + 1000;

    while ( ! isStartPressed()) {
      // process the time increment timerVal & display 1 sec intervals
      if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal++;

            if( timerVal >= 5 * 60) return;
          
            // show the time
            showTime(minsecfmt, 4);
       }
   }  // while   
}  // displayMedAlarm()

// Meditation recovery chirp then continue count down till start pressed
void displayMedRecover() {
   long targetTm;
   tone(SPKER, 600, 1000);  // 1 sec chirp
   timerVal = 5 * 60 + 1;  // allow for chirp time

   showTime(minsecfmt, 4);
   targetTm = millis() + 1000;

    while ( ! isStartPressed()) {
      // process the time increment timerVal & display 1 sec intervals
      if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal++;

            // show the time
            showTime(minsecfmt, 4);
       }
    }  // while   
}  // displayMedRecover()

int displayStdRange(){
   startDisplay();
   u8x8.setCursor(0,2);  // set position of menu title
   myMenu.initial("  Timer Range? ", 2, 0, 3); // menu title, # menu items, initial selected
                                              // starting at line
   myMenu.addMenuSel(0, 2, "Min/Sec" );  // line, col, menu text
   myMenu.addMenuSel(1, 2, "Hr/Min" ); 

  
   myMenu.showMenu(); // show items 

   return getMenuSel();  
}  // displayStdRange()

// allows selecton of a free form timer value either minutes and
// seconds or hours and minites depending on minsecfmt
// pressing encoder button will change encoder input from major to
// minor value input, pressing start will set the timerVal variable
// and exit
void displayStdSel() {
   int tval;  //  raw timer value assume seconds, if hr format 
              //  will need conversion.
         
   startDisplay();
   u8x8.setCursor(0,2);  // set position of menu title
   u8x8.print("  Set the Time ");

   u8x8.setCursor(0,7);  // set position of menu title
   if( minsecfmt) {
      u8x8.print("Minutes:Seconds"); 
   }
   else {
      u8x8.print(" Hours:Minutes "); 
   }

   // display initial time "00:00"
   tval = getFreeTime();
   if( ! minsecfmt) tval *= 60;

   timerVal = tval;
}  //  displayStdSel()

// returns tnhe value of the alarm type 
// None = 0, audible = 1, relayctl = 2, Both = 3
byte displayStdCtl() {
   startDisplay();
   u8x8.setCursor(0,2);  // set position of menu title
   myMenu.initial("  Alarm Type? ", 4, 2, 3); // menu title, # menu items, initial selected
                                              // starting at line
   myMenu.addMenuSel(0, 2, "None" );  // line, col, menu text
   myMenu.addMenuSel(1, 2, "Audible" ); 
   myMenu.addMenuSel(2, 2, "Control" );  // line, col, menu text
   myMenu.addMenuSel(3, 2, "Both" ); 

  
   myMenu.showMenu(); // show items 

   return getMenuSel();     
}  // displayStdCtl()

// display the selected time and wait for start button pressed
void displayStdTime(){
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print(" Selected Time");
    showTime(minsecfmt, 4);
    u8x8.setCursor(0,7);  // show start msg
    u8x8.print("Press START ->");

    // wait for start pressed
    while(! isStartPressed()) {
      
    }
       
}  // displayStdTime()

// display count down till zero
int displayStdCountdn(){
    long targetTm;
    startDisplay();
    u8x8.setCursor(0,2);  // show time title
    u8x8.print(" Time Remaining ");
    showTime(minsecfmt, 4);
    // if control selected turn on relay
    if ( (alarmType & CONTROL) != 0) digitalWrite(RELAY, RLYON);
    
    targetTm = millis() + 1000;
     while( true) {
         if( isStartPressed()) return -1;  // abort
         if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal--;
            if( timerVal <= 0) return 0;  // normal function return
            // test for minute update
            if( ! minsecfmt && (timerVal % 60) != 0) continue;  // while loop 

            // show the time
            showTime(minsecfmt, 4);
         }
      
     }  // while
     // should never get here!
     Serial.println(F("ERROR - in displayStdCountdn()"));       
}  //displayStdCountdn()

void displayStdAlarm() {
    long targetTm;
   digitalWrite(RELAY, RLYOFF);  // doesn't hurt to turn off if already off
   if( (alarmType & AUDIBLE) != 0) {  //  Audible alarm
      tone(SPKER, 440);
   }

   // show over time
   startDisplay();
   u8x8.setCursor(0,2);  // show time title
   u8x8.print("  Timer Over   ");
   u8x8.setInverseFont(1);
   showTime(minsecfmt, 4);
   targetTm = millis() + 1000;

   while ( ! isStartPressed()) {
      // process the time increment timerVal & display 1 sec intervals
      if( millis() >= targetTm) {
            targetTm = millis() + 1000;  // reset second timer
            timerVal++;
          
            // show the time
            showTime(minsecfmt, 4);
       }
   }  // while
      
   u8x8.setInverseFont(0);
}  // displayStdAlarm()
